import os
from PyQt6 import uic
from PyQt6.QtWidgets import QWidget, QVBoxLayout, QSpacerItem, QSizePolicy
from PyQt6.QtGui import QIcon


class LayerMixin:
    """
    This is a mixin class for the main window that handles the layer manager.
    """
    @property
    def layer_highlight_style(self):
        return f"background-color: #c7f6fc; border: 1px solid #000000; color: #000000;"

    @property
    def layer_normal_style(self):
        return "background-color: #ffffff; border: 1px solid #333333; color: #333;"

    def initialize(self):
        self.window.new_layer.clicked.connect(self.new_layer)
        self.window.layer_up_button.clicked.connect(self.layer_up_button)
        self.window.layer_down_button.clicked.connect(self.layer_down_button)
        self.window.delete_layer_button.clicked.connect(self.delete_layer_button)

    def layer_up_button(self):
        self.canvas.move_layer_up(self.canvas.current_layer)
        self.show_layers()

    def layer_down_button(self):
        self.canvas.move_layer_down(self.canvas.current_layer)
        self.show_layers()

    def delete_layer_button(self):
        self.canvas.delete_layer(self.canvas.current_layer_index)
        self.show_layers()

    def new_layer(self):
        self.canvas.add_layer()
        self.show_layers()

    def show_layers(self):
        """
        This function is called when the layers need to be updated.
        :return:
        """

        # create an object which can contain a layer_obj and then be added to layers.setWidget
        container = QWidget()
        container.setLayout(QVBoxLayout())

        for layer in self.canvas.layers:
            index = self.canvas.layers.index(layer)
            layer_obj = uic.loadUi(os.path.join("pyqt/layer.ui"))
            layer_obj.layer_name.setText(layer.name)

            layer_obj.opacity_slider.valueChanged.connect(
                lambda val, _layer=layer_obj, _index=index: self.slider_set_layer_opacity(val, _layer, _index))
            layer_obj.opacity_spinbox.valueChanged.connect(
                lambda val, _layer=layer_obj, _index=index: self.spinbox_set_layer_opacity(val, _layer, _index))
            opacity = self.canvas.get_layer_opacity(index)
            layer_obj.opacity_slider.setValue(int(opacity * 100))
            layer_obj.opacity_spinbox.setValue(opacity)

            # onclick of layer_obj set as the current layer index on self.canvas
            layer_obj.mousePressEvent = lambda event, _layer=layer, _index=index: self.set_current_layer(_index)

            # show a border around layer_obj if it is the selected index
            if self.canvas.current_layer_index == index:
                layer_obj.frame.setStyleSheet(self.layer_highlight_style)
            else:
                layer_obj.frame.setStyleSheet(self.layer_normal_style)

            layer_obj.visible_button.setIcon(QIcon("src/icons/eye.png" if layer.visible else "src/icons/eye-off.png"))
            layer_obj.visible_button.clicked.connect(
                lambda _, _layer=layer, _layer_obj=layer_obj: self.toggle_layer_visibility(_layer, _layer_obj))

            container.layout().addWidget(layer_obj)

        # add a spacer to the bottom of the container
        container.layout().addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding))

        self.window.layers.setWidget(container)
        self.container = container

    def slider_set_layer_opacity(self, val, layer_obj, index):
        val = val / 100
        self.canvas.set_layer_opacity(index, val)
        layer_obj.opacity_spinbox.setValue(val)
        self.canvas.layers[index].opacity = val

    def spinbox_set_layer_opacity(self, val, layer_obj, index):
        self.canvas.set_layer_opacity(index, val)
        val = int(val * 100)
        layer_obj.opacity_slider.setValue(val)
        self.canvas.layers[index].opacity = val

    def toggle_layer_visibility(self, layer, layer_obj):
        # change the eye icon of the visible_button on the layer
        self.canvas.toggle_layer_visibility(layer)
        layer_obj.visible_button.setIcon(QIcon("src/icons/eye.png" if layer.visible else "src/icons/eye-off.png"))

    def set_current_layer(self, index):
        if not hasattr(self, "container"):
            return
        if self.container:
            item = self.container.layout().itemAt(self.canvas.current_layer_index)
            if item:
                item.widget().frame.setStyleSheet(self.layer_normal_style)
        self.canvas.current_layer_index = index
        if self.container:
            item = self.container.layout().itemAt(self.canvas.current_layer_index)
            if item:
                item.widget().frame.setStyleSheet(self.layer_highlight_style)

    def delete_layer(self):
        pass



import os
import random
from PIL import Image
from PyQt6 import uic
from PyQt6.QtCore import QRect, pyqtSignal
from PyQt6.uic.exceptions import UIFileException
from aihandler.settings import MAX_SEED, AVAILABLE_SCHEDULERS_BY_ACTION, MODELS
from airunner.windows.video import VideoPopup
from airunner.utils import load_default_models, load_models_from_path
from airunner.mixins.lora_mixin import LoraMixin
from PIL import PngImagePlugin


class GeneratorMixin(LoraMixin):
    generate_signal = pyqtSignal(dict)

    @property
    def available_models(self):
        return MODELS

    @property
    def model_base_path(self):
        return self.settings.model_base_path.get()

    @model_base_path.setter
    def model_base_path(self, val):
        self.settings.model_base_path.set(val)

    @property
    def width(self):
        return int(self.settings_manager.settings.working_width.get())

    @width.setter
    def width(self, val):
        self.settings_manager.settings.working_width.set(val)
        self.set_final_size_label()
        self.canvas.update()

    @property
    def height(self):
        return int(self.settings_manager.settings.working_height.get())

    @height.setter
    def height(self, val):
        self.settings_manager.settings.working_height.set(val)
        self.set_final_size_label()
        self.canvas.update()

    @property
    def steps(self):
        return self.settings.steps.get()

    @steps.setter
    def steps(self, val):
        self.settings.steps.set(val)

    @property
    def ddim_eta(self):
        return self.settings.ddim_eta.get()

    @ddim_eta.setter
    def ddim_eta(self, val):
        self.settings.ddim_eta.set(val)

    @property
    def prompt(self):
        return self.settings.prompt.get()

    @prompt.setter
    def prompt(self, val):
        self.settings.prompt.set(val)

    @property
    def negative_prompt(self):
        return self.settings.negative_prompt.get()

    @negative_prompt.setter
    def negative_prompt(self, val):
        self.settings.negative_prompt.set(val)

    @property
    def scale(self):
        return self.settings.scale.get()

    @scale.setter
    def scale(self, val):
        self.settings.scale.set(val)

    @property
    def image_scale(self):
        return self.settings.image_guidance_scale.get()

    @image_scale.setter
    def image_scale(self, val):
        self.settings.image_guidance_scale.set(val)

    @property
    def strength(self):
        return self.settings.strength.get()

    @strength.setter
    def strength(self, val):
        self.settings.strength.set(val)

    @property
    def seed(self):
        return self.settings.seed.get()

    @seed.setter
    def seed(self, val):
        self.settings.seed.set(val)

    @property
    def random_seed(self):
        return self.settings.random_seed.get()

    @random_seed.setter
    def random_seed(self, val):
        self.settings.random_seed.set(val)

    @property
    def samples(self):
        return self.settings.n_samples.get()

    @samples.setter
    def samples(self, val):
        self.settings.n_samples.set(val)

    @property
    def model(self):
        return self.settings.model_var.get()

    @model.setter
    def model(self, val):
        self.settings.model_var.set(val)

    @property
    def scheduler(self):
        return self.settings.scheduler_var.get()

    @scheduler.setter
    def scheduler(self, val):
        self.settings.scheduler_var.set(val)

    @property
    def downscale_amount(self):
        return self.settings.downscale_amount.get()

    @downscale_amount.setter
    def downscale_amount(self, val):
        self.settings.downscale_amount.set(val)

    @property
    def do_upscale_by_active_grid(self):
        return self.settings.do_upscale_by_active_grid.get()

    @do_upscale_by_active_grid.setter
    def do_upscale_by_active_grid(self, val):
        self.settings.do_upscale_by_active_grid.set(val)

    @property
    def do_upscale_full_image(self):
        return self.settings.do_upscale_full_image.get()

    @do_upscale_full_image.setter
    def do_upscale_full_image(self, val):
        self.settings.do_upscale_full_image.set(val)

    @property
    def image_to_new_layer(self):
        return self.settings_manager.settings.image_to_new_layer.get()

    def update_prompt(self, prompt):
        self.tabs[self.current_section].prompt.setPlainText(prompt)

    def update_negative_prompt(self, prompt):
        self.tabs[self.current_section].negative_prompt.setPlainText(prompt)

    def initialize(self):
        # create an emitter that will be triggered when generate is called
        self.settings_manager.settings.model_base_path.my_signal.connect(self.refresh_model_list)
        for tab_section in self._tabs.keys():  # iterate over each tab section (stablediffusion, kandinsky)
            self.override_tab_section = tab_section
            for tab in self.tabs.keys():  # iterate over each section within the tab section (txt2img, img2img, etc)
                self.tabs[tab] = uic.loadUi(os.path.join("pyqt/generate_form.ui"))

                # Remove embeddings and LoRA for Kandinsky sections as they are not supported
                if tab_section == "kandinsky":
                    self.tabs[tab].PromptTabsSection.removeTab(2)
                    self.tabs[tab].PromptTabsSection.removeTab(1)

                # set up the override_section for the tab so that we can force which tab is active
                # this is a hack in order to get around the auto-switching of tabs when the user
                # changes the section
                override_section = tab
                self.override_section = override_section

                # Initialize controlnet dropdown for controlnet section
                if tab == "controlnet":
                    controlnet_options = [
                        "Canny",
                        "MLSD",
                        "Depth",
                        "Normal",
                        "Segmentation",
                        "Lineart",
                        "Openpose",
                        "Scribble",
                        "Softedge",
                        "Pixel2Pixel",
                        "Inpaint",
                        "Shuffle",
                        "Anime",
                    ]
                    for option in controlnet_options:
                        self.tabs[tab].controlnet_dropdown.addItem(option)

                """
                The generator widget contains many different settings, however not all settings are
                available to each section. Rather than creating a new widget for each section, it is
                easier to delete the settings that are not available for a given section.
                """
                # delete strength slider for given sections
                if tab in ["txt2img", "pix2pix", "outpaint", "upscale", "superresolution", "txt2vid"]:
                    self.tabs[tab].strength.deleteLater()

                # delete image scale box for given sections
                if tab in ["txt2img", "img2img", "depth2img", "outpaint", "controlnet", "superresolution", "txt2vid"]:
                    self.tabs[tab].image_scale_box.deleteLater()

                # delete the sample slider for given sections
                if tab in ["upscale", "superresolution"]:
                    self.tabs[tab].samples_groupbox.deleteLater()
                    self.tabs[tab].full_image_radiobutton.setChecked(self.do_upscale_full_image == True)
                    self.tabs[tab].active_grid_radiobutton.setChecked(self.do_upscale_by_active_grid == True)
                    # handle radiobutton change
                    self.tabs[tab].full_image_radiobutton.toggled.connect(
                        lambda val, _tab=self.tabs[tab]: self.handle_upscale_full_image_change(val, _tab)
                    )
                    self.tabs[tab].active_grid_radiobutton.toggled.connect(
                        lambda val, _tab=self.tabs[tab]: self.handle_upscale_active_grid_change(val, _tab)
                    )
                    self.tabs[tab].downscale_spinbox.setValue(self.downscale_amount)
                    self.tabs[tab].downscale_spinbox.valueChanged.connect(
                        lambda val, _tab=self.tabs[tab]: self.handle_downscale_spinbox_change(val, _tab)
                    )
                    self.set_final_size_label(self.tabs[tab])
                else:
                    # delete downscale settings for sections that are not upscale or superresolution
                    self.tabs[tab].downscale_settings.deleteLater()
                    self.tabs[tab].downscale_settings_radiobuttons.deleteLater()
                    self.tabs[tab].downscale_settings_sizes.deleteLater()
                    self.tabs[tab].active_grid_radiobutton.deleteLater()
                    self.tabs[tab].full_image_radiobutton.deleteLater()
                    self.tabs[tab].downscale_label.deleteLater()
                    self.tabs[tab].downscale_amount_container.deleteLater()
                    self.tabs[tab].downscale_spinbox.deleteLater()
                    self.tabs[tab].final_size_label.deleteLater()
                    self.tabs[tab].final_size.deleteLater()

                # delete controlnet settings for sections that are not controlnet
                if tab != "controlnet":
                    self.tabs[tab].controlnet_label.deleteLater()
                    self.tabs[tab].controlnet_dropdown.deleteLater()

                self.tabs[tab].interrupt_button.clicked.connect(self.interrupt)

        self.override_section = None

        for tab_section in self._tabs.keys():
            self.override_tab_section = tab_section
            for tab in self.tabs.keys():
                display_name = tab
                if display_name == "outpaint":
                    display_name = "inpaint / outpaint"

                if tab_section == "stablediffusion":
                    self.window.stableDiffusionTabWidget.addTab(self.tabs[tab], display_name)
                else:
                    self.window.kandinskyTabWidget.addTab(self.tabs[tab], display_name)

                self.tabs[tab].generate.clicked.connect(self.generate_callback)

            # iterate over each tab and connect steps_slider with steps_spinbox
            for tab_name in self.tabs.keys():
                self.override_section = tab_name
                tab = self.tabs[tab_name]

                tab.toggleAllLora.clicked.connect(lambda checked, _tab=tab: self.toggle_all_lora(checked, _tab))

                # tab.prompt is QPlainTextEdit - on text change, call handle_prompt_change
                tab.prompt.textChanged.connect(lambda _tab=tab: self.handle_prompt_change(_tab))
                tab.negative_prompt.textChanged.connect(lambda _tab=tab: self.handle_negative_prompt_change(_tab))

                tab.steps_slider.valueChanged.connect(lambda val, _tab=tab: self.handle_steps_slider_change(val, _tab))
                tab.steps_spinbox.valueChanged.connect(lambda val, _tab=tab: self.handle_steps_spinbox_change(val, _tab))

                # load models by section
                self.load_model_by_section(tab_section, tab, tab_name)

                # on change of tab.model_dropdown set the model in self.settings_manager
                tab.model_dropdown.currentIndexChanged.connect(
                    lambda val, _tab=tab, _section=tab_name: self.set_model(_tab, _section, val)
                )

                # set schedulers for each tab
                scheduler_action = tab_name
                if tab_section == "kandinsky":
                    scheduler_action = f"kandinsky_{tab_name}"
                tab.scheduler_dropdown.addItems(AVAILABLE_SCHEDULERS_BY_ACTION[scheduler_action])
                tab.scheduler_dropdown.currentIndexChanged.connect(
                    lambda val, _tab=tab, _section=tab_name: self.set_scheduler(_tab, _section, val)
                )

                # scale slider
                tab.scale_slider.valueChanged.connect(lambda val, _tab=tab: self.handle_scale_slider_change(val, _tab))
                tab.scale_spinbox.valueChanged.connect(lambda val, _tab=tab: self.handle_scale_spinbox_change(val, _tab))

                tab.image_scale_slider.valueChanged.connect(
                    lambda val, _tab=tab: self.handle_image_scale_slider_change(val, _tab))
                tab.image_scale_spinbox.valueChanged.connect(
                    lambda val, _tab=tab: self.handle_image_scale_spinbox_change(val, _tab))

                # strength slider
                section = tab_name
                strength = 0
                if section in ["img2img", "depth2img", "controlnet"]:
                    if section == "img2img":
                        strength = self.settings_manager.settings.img2img_strength.get()
                    elif section == "depth2img":
                        strength = self.settings_manager.settings.depth2img_strength.get()
                    elif section == "controlnet":
                        strength = self.settings_manager.settings.controlnet_strength.get()
                    tab.strength_slider.setValue(int(strength))
                    tab.strength_spinbox.setValue(strength / 100)
                    tab.strength_slider.valueChanged.connect(
                        lambda val, _tab=tab: self.handle_strength_slider_change(val, _tab))
                    tab.strength_spinbox.valueChanged.connect(
                        lambda val, _tab=tab: self.handle_strength_spinbox_change(val, _tab))

                if section == "txt2vid":
                    # change the label tab.samples_groupbox label to "Frames"
                    tab.samples_groupbox.setTitle("Frames")

                tab.seed.textChanged.connect(lambda _tab=tab: self.text_changed(_tab))
                tab.random_checkbox.stateChanged.connect(
                    lambda val, _tab=tab: self.handle_random_checkbox_change(val, _tab))

                tab.random_checkbox.setChecked(self.random_seed is True)

                # samples slider
                tab.samples_slider.valueChanged.connect(
                    lambda val, _tab=tab: self.handle_samples_slider_change(val, _tab))
                tab.samples_spinbox.valueChanged.connect(
                    lambda val, _tab=tab: self.handle_samples_spinbox_change(val, _tab))

                # if samples is greater than 1 enable the interrupt_button
                if tab.samples_spinbox.value() > 1:
                    tab.interrupt_button.setEnabled(tab.samples_spinbox.value() > 1)
                self.set_default_values(tab_name, tab)
                self.override_section = None

            # assign callback to generate function on tab
            if tab_section == "stablediffusion":
                self.window.stableDiffusionTabWidget.currentChanged.connect(self.tab_changed_callback)
            else:
                self.window.kandinskyTabWidget.currentChanged.connect(self.tab_changed_callback)

        self.override_tab_section = None

        self.initialize_size_form_elements()
        self.initialize_size_sliders()
        self.initialize_lora()

    def handle_downscale_spinbox_change(self, value, tab):
        current_value = self.downscale_amount
        new_value = value
        if value == 1:
            if value > current_value or value == current_value:
                new_value = 2
            else:
                new_value = 0
        self.downscale_amount = value
        tab.downscale_spinbox.setValue(new_value)
        self.set_final_size_label(tab)

    def set_final_size_label(self, tab=None):
        if tab is None:
            tab = self.tabs[self.current_section]

        if self.current_section not in ["upscale", "superresolution"]:
            return

        image = self.canvas.current_layer.image_data.image
        if image:
            if self.do_upscale_by_active_grid:
                width = self.settings_manager.settings.working_width.get()
                height = self.settings_manager.settings.working_height.get()
            else:
                width = image.width
                height = image.height
        else:
            width = 0
            height = 0

        if self.downscale_amount > 0:
            width = width // self.downscale_amount
            height = height // self.downscale_amount

        if self.current_section == "upscale":
            width = width * 2
            height = height * 2
        elif self.current_section == "superresolution":
            width = width * 4
            height = height * 4

        # set final_size label text
        tab.final_size.setText(f"{width}x{height}")

    def handle_upscale_full_image_change(self, value, tab):
        self.do_upscale_full_image = value
        self.do_upscale_by_active_grid = value == False
        self.set_final_size_label(tab)

    def handle_upscale_active_grid_change(self, value, tab):
        self.do_upscale_by_active_grid = value
        self.do_upscale_full_image = value == False
        self.set_final_size_label(tab)

    def interrupt(self):
        print("Interrupting...")
        self.client.sd_runner.cancel()

    def refresh_model_list(self):
        for i, section in enumerate(self._tabs[self.currentTabSection].keys()):
            tab = self.tabWidget.widget(i)
            self.clear_model_list(tab)
            self.load_model_by_section(self.currentTabSection, tab, section)

    def clear_model_list(self, tab):
        tab.model_dropdown.clear()

    def load_model_by_section(self, tab_section, tab, section_name):
        if section_name in ["txt2img", "img2img"]:
            section_name = "generate"

        models = self.models if self.models else []
        default_models = load_default_models(tab_section, section_name)
        path = ""
        if section_name == "depth2img":
            path = self.settings_manager.settings.depth2img_model_path.get()
        elif section_name == "pix2pix":
            path = self.settings_manager.settings.pix2pix_model_path.get()
        elif section_name == "outpaint":
            path = self.settings_manager.settings.outpaint_model_path.get()
        elif section_name == "upscale":
            path = self.settings_manager.settings.upscale_model_path.get()
        if not path or path == "":
            path = self.settings_manager.settings.model_base_path.get()

        if tab_section == "stablediffusion":
            new_models = load_models_from_path(path)
            default_models += new_models
        models += default_models
        self.models = models

        tab.model_dropdown.addItems(default_models)

    def reset_settings(self):
        self.settings_manager.reset_settings_to_default()
        for tab_name in self.tabs.keys():
            tab = self.tabs[tab_name]
            self.set_default_values(tab_name, tab)
        self.canvas.update()


    def text_changed(self, tab):
        try:
            val = int(tab.seed.toPlainText())
            self.seed = val
        except ValueError:
            pass

    def handle_random_checkbox_change(self, val, _tab):
        if val == 2:
            self.random_seed = True
        else:
            self.random_seed = False
        _tab.seed.setEnabled(not self.random_seed)

    def video_handler(self, data):
        filename = data["video_filename"]
        VideoPopup(settings_manager=self.settings_manager, file_path=filename)

    def image_handler(self, image, data, nsfw_content_detected):
        if data["action"] == "txt2vid":
            return self.video_handler(data)

        if self.settings_manager.settings.auto_export_images.get():
            self.auto_export_image(image, data)

        self.stop_progress_bar(data["action"])
        if nsfw_content_detected and self.settings_manager.settings.nsfw_filter.get():
            self.message_handler("NSFW content detected, try again.", error=True)
        else:
            if data["action"] != "outpaint" and self.image_to_new_layer and self.canvas.current_layer.image_data.image is not None:
                self.canvas.add_layer()
            # print width and height of image
            self.canvas.image_handler(image, data)
            self.message_handler("")
            self.show_layers()

    def load_metadata(self, metadata):
        if metadata:
            action = metadata.get("action")
            prompt = None
            negative_prompt = None
            if "prompt" in metadata:
                prompt = metadata.get("prompt", "")
            if "negative_prompt" in metadata:
                negative_prompt = metadata.get("negative_prompt", "")
            scale = metadata.get("scale", None)
            seed = metadata.get("seed", None)
            steps = metadata.get("steps", None)
            ddim_eta = metadata.get("ddim_eta", None)
            n_iter = metadata.get("n_iter", None)
            n_samples = metadata.get("n_samples", None)
            model = metadata.get("model", None)
            # model_branch = metadata.get("model_branch", None)
            scheduler = metadata.get("scheduler", None)
            if prompt is not None:
                self.tabs[action].prompt.setPlainText(prompt)
            if negative_prompt is not None:
                self.tabs[action].negative_prompt.setPlainText(negative_prompt)
            if scale:
                scale = float(scale)
                self.tabs[action].scale_spinbox.setValue(float(scale))
                self.tabs[action].scale_slider.setValue(int(float(scale) * 100))
            if seed:
                self.tabs[action].seed.setPlainText(seed)
            if steps:
                steps = int(steps)
                self.tabs[action].steps_spinbox.setValue(steps)
                self.tabs[action].steps_slider.setValue(steps)
            if ddim_eta:
                ddim_eta = float(ddim_eta)
                self.tabs[action].ddim_eta_spinbox.setValue(ddim_eta)
                self.tabs[action].ddim_eta_slider.setValue(ddim_eta * 100)
            if n_iter:
                n_iter = int(n_iter)
                try:
                    self.tabs[action].n_iter_spinbox.setValue(n_iter)
                except AttributeError:
                    pass
                try:
                    self.tabs[action].n_iter_slider.setValue(n_iter)
                except AttributeError:
                    pass
            if n_samples:
                n_samples = int(n_samples)
                self.tabs[action].samples_spinbox.setValue(n_samples)
                self.tabs[action].samples_slider.setValue(n_samples)
            if model:
                self.tabs[action].model_dropdown.setCurrentText(model)
            if scheduler:
                self.tabs[action].scheduler_dropdown.setCurrentText(scheduler)

    def prepare_metadata(self, data):
        if not self.settings_manager.settings.export_metadata.get() or \
                self.settings_manager.settings.image_export_type.get() != "png":
            return None
        metadata = PngImagePlugin.PngInfo()
        options = data["options"]
        action = data["action"]
        metadata.add_text("action", action)
        if self.settings_manager.settings.image_export_metadata_prompt.get() is True:
            metadata.add_text("prompt", options[f'{action}_prompt'])
        if self.settings_manager.settings.image_export_metadata_negative_prompt.get() is True:
            metadata.add_text("negative_prompt", options[f'{action}_negative_prompt'])
        if self.settings_manager.settings.image_export_metadata_scale.get() is True:
            metadata.add_text("scale", str(options[f"{action}_scale"]))
        if self.settings_manager.settings.image_export_metadata_seed.get() is True:
            metadata.add_text("seed", str(options[f"{action}_seed"]))
        if self.settings_manager.settings.image_export_metadata_steps.get() is True:
            metadata.add_text("steps", str(options[f"{action}_steps"]))
        if self.settings_manager.settings.image_export_metadata_ddim_eta.get() is True:
            metadata.add_text("ddim_eta", str(options[f"{action}_ddim_eta"]))
        if self.settings_manager.settings.image_export_metadata_iterations.get() is True:
            metadata.add_text("n_iter", str(options[f"{action}_n_iter"]))
        if self.settings_manager.settings.image_export_metadata_samples.get() is True:
            metadata.add_text("n_samples", str(options[f"{action}_n_samples"]))
        if self.settings_manager.settings.image_export_metadata_model.get() is True:
            metadata.add_text("model", str(options[f"{action}_model"]))
        if self.settings_manager.settings.image_export_metadata_model_branch.get() is True:
            metadata.add_text("model_branch", str(options[f"{action}_model_branch"]))
        if self.settings_manager.settings.image_export_metadata_scheduler.get() is True:
            metadata.add_text("scheduler", str(options[f"{action}_scheduler"]))
        return metadata

    def auto_export_image(self, image, data):
        """
        Export image along with stats to image_path
        :return:
        """
        if data["action"] == "txt2vid":
            return
        base_path = self.settings_manager.settings.model_base_path.get()
        image_path = self.settings_manager.settings.image_path.get()
        image_path = "images" if image_path == "" else image_path
        path = os.path.join(base_path, image_path) if image_path == "images" else image_path
        if not os.path.exists(path):
            os.makedirs(path)
        # check for existing files, if they exist, increment the filename. filename should be in the format
        # <action>_<seed>_<N>.png
        extension = f".{self.settings_manager.settings.image_export_type.get()}"
        filename = data["action"] + "_" + str(self.seed)
        if os.path.exists(os.path.join(path, filename + extension)):
            i = 1
            while os.path.exists(os.path.join(path, filename + "_" + str(i) + extension)):
                i += 1
            filename = filename + "_" + str(i)
        metadata = self.prepare_metadata(data)
        if metadata:
            image.save(os.path.join(path, filename + extension), pnginfo=metadata)
        else:
            image.save(os.path.join(path, filename + extension))

    def handle_steps_slider_change(self, val, tab):
        tab.steps_spinbox.setValue(int(val))
        self.steps = int(val)

    def handle_prompt_change(self, tab):
        self.prompt = tab.prompt.toPlainText()

    def handle_negative_prompt_change(self, tab):
        self.negative_prompt = tab.negative_prompt.toPlainText()

    def handle_steps_spinbox_change(self, val, tab):
        tab.steps_slider.setValue(int(val))
        self.steps = int(val)

    def handle_scale_slider_change(self, val, tab):
        tab.scale_spinbox.setValue(val / 100.0)
        self.scale = val

    def handle_image_scale_slider_change(self, val, tab):
        tab.image_scale_spinbox.setValue(val / 100.0)
        try:
            self.image_scale = val
        except:
            pass

    def handle_image_scale_spinbox_change(self, val, tab):
        tab.image_scale_slider.setValue(int(val * 100))
        try:
            self.image_scale = val * 100
        except:
            pass

    def handle_scale_spinbox_change(self, val, tab):
        tab.scale_slider.setValue(int(val * 100))
        self.scale = val * 100

    def handle_strength_slider_change(self, val, tab):
        tab.strength_spinbox.setValue(val / 100.0)
        self.strength = val

    def handle_strength_spinbox_change(self, val, tab):
        tab.strength_slider.setValue(int(val * 100))
        self.strength = val

    def handle_seed_spinbox_change(self, val, tab):
        tab.seed.setText(str(int(val)))
        self.seed = int(val)

    def handle_samples_slider_change(self, val, tab):
        tab.samples_spinbox.setValue(int(val))
        self.samples = int(val)
        tab.interrupt_button.setEnabled(tab.samples_spinbox.value() > 1)

    def handle_samples_spinbox_change(self, val, tab):
        tab.samples_slider.setValue(int(val))
        self.samples = int(val)
        tab.interrupt_button.setEnabled(tab.samples_spinbox.value() > 1)

    def set_model(self, tab, section, val):
        model = tab.model_dropdown.currentText()
        self.model = model

    def set_scheduler(self, tab, section, val):
        scheduler = tab.scheduler_dropdown.currentText()
        self.scheduler = scheduler

    def generate_callback(self):
        # check that the correct model is in use for txt2vid
        if self.current_section == "txt2vid" and self.tabs[self.current_section].model_dropdown.currentText() != "damo-vilab":
            self.error_handler("Must use damo-vilab model with txt2vid")
        else:
            self.generate(True)


    def prep_video(self):
        pass

    @property
    def active_rect(self):
        rect = QRect(
            self.canvas.active_grid_area_rect.x(),
            self.canvas.active_grid_area_rect.y(),
            self.canvas.active_grid_area_rect.width(),
            self.canvas.active_grid_area_rect.height(),
        )
        rect.translate(-self.canvas.pos_x, -self.canvas.pos_y)

        return rect

    def tab_has_embeddings(self, tab):
        return tab not in ["upscale", "superresolution", "txt2vid"]

    def generate(self, image=None):
        if self.current_section in ("upscale", "superresolution") and self.do_upscale_full_image:
            image_data = self.canvas.current_layer.image_data
            image = image_data.image if image_data else None
            if image is None:
                self.error_handler("No image to upscale")
                return
            downscale_amount = self.downscale_amount
            if downscale_amount > 0:
                # downscale the image first
                image = image.resize(
                    (
                        int(image.width // downscale_amount),
                        int(image.height // downscale_amount),
                    ),
                    Image.BICUBIC,
                )
            self.requested_image = image
            self.start_progress_bar(self.current_section)
            self.do_generate({
                "mask": image.convert("RGB"),
                "image": image.convert("RGB"),
                "location": image_data.position
            })
        elif self.use_pixels:
            self.requested_image = image
            self.start_progress_bar(self.current_section)
            image_data = self.canvas.current_layer.image_data
            image = image_data.image if image_data else None

            if image is None:
                # create a transparent image the size of self.canvas.active_grid_area_rect
                width = self.settings_manager.settings.working_width.get()
                height = self.settings_manager.settings.working_height.get()
                image = Image.new("RGBA", (int(width), int(height)), (0, 0, 0, 0))

            img = image.copy().convert("RGBA")
            new_image = Image.new(
                "RGBA",
                (self.settings.working_width.get(), self.settings.working_height.get()),
                (0, 0, 0))

            cropped_outpaint_box_rect = self.active_rect
            crop_location = (
                cropped_outpaint_box_rect.x() - self.canvas.image_pivot_point.x(),
                cropped_outpaint_box_rect.y() - self.canvas.image_pivot_point.y(),
                cropped_outpaint_box_rect.width() - self.canvas.image_pivot_point.x(),
                cropped_outpaint_box_rect.height() - self.canvas.image_pivot_point.y()
            )
            new_image.paste(img.crop(crop_location), (0, 0))
            # save new_image to disc
            mask = Image.new("RGB", (new_image.width, new_image.height), (255, 255, 255))
            for x in range(new_image.width):
                for y in range(new_image.height):
                    try:
                        if new_image.getpixel((x, y))[3] != 0:
                            mask.putpixel((x, y), (0, 0, 0))
                    except IndexError:
                        pass

            # convert image to rgb
            image = new_image.convert("RGB")
            self.do_generate({
                "mask": mask,
                "image": image,
                "location": self.canvas.active_grid_area_rect
            })
        elif self.action == "vid2vid":
            images = self.prep_video()
            self.do_generate({
                "images": images
            })
        else:
            self.do_generate()

    def start_progress_bar(self, section):
        # progressBar: QProgressBar = self.tabs[section].progressBar
        # progressBar.setRange(0, 0)
        if self.progress_bar_started:
            return
        self.progress_bar_started = True
        self.tqdm_callback_triggered = False
        self.stop_progress_bar(section)
        self.tabs[section].progressBar.setRange(0, 0)
        self.tqdm_var.set({
            "step": 0,
            "total": 0,
            "action": section,
            "image": None,
            "data": None
        })

    def stop_progress_bar(self, section):
        self.tabs[section].progressBar.reset()
        self.tabs[section].progressBar.setRange(0, 100)

    def do_generate(self, extra_options=None):
        if not extra_options:
            extra_options = {}

        # self.start_progress_bar(self.current_section)

        action = self.current_section
        tab = self.tabs[action]

        if self.random_seed:
            # randomize seed
            seed = random.randint(0, MAX_SEED)
            self.seed = seed
            # set random_seed on current tab
            self.tabs[action].seed.setText(str(seed))
        if action in ("txt2img", "img2img", "pix2pix", "depth2img", "txt2vid"):
            samples = self.samples
        else:
            samples = 1

        prompt = self.tabs[action].prompt.toPlainText()
        negative_prompt = self.tabs[action].negative_prompt.toPlainText()
        if self.random_seed:
            seed = random.randint(0, MAX_SEED)
            self.settings.seed.set(seed)
        else:
            seed = self.seed

        # set the model data
        model = tab.model_dropdown.currentText()
        model_branch = None
        section_name = action
        if section_name in ["txt2img", "img2img"]:
            section_name = "generate"

        section_name = f"{self.currentTabSection}_{section_name}"

        if model in self.available_models[section_name]:
            model_path = self.available_models[section_name][model]["path"]
            model_branch = self.available_models[section_name][model].get("branch", "main")
        elif model not in self.models:
            model_names = list(self.available_models[section_name].keys())
            model = model_names[0]
            model_path = self.available_models[section_name][model]["path"]
            model_branch = self.available_models[section_name][model].get("branch", "main")
        else:
            path = self.settings_manager.settings.model_base_path.get()
            if action == "depth2img":
                path = self.settings_manager.settings.depth2img_model_path.get()
            elif action == "pix2pix":
                path = self.settings_manager.settings.pix2pix_model_path.get()
            elif action == "outpaint":
                path = self.settings_manager.settings.outpaint_model_path.get()
            elif action == "upscale":
                path = self.settings_manager.settings.upscale_model_path.get()
            model_path = os.path.join(path, model)

        # get controlnet_dropdown from active tab
        use_controlnet = False
        controlnet = ""
        if action == "controlnet":
            controlnet_dropdown = self.tabs[action].controlnet_dropdown
            # get controlnet from controlnet_dropdown
            controlnet = controlnet_dropdown.currentText()
            controlnet = controlnet.lower()
            use_controlnet = controlnet != "none"

        options = {
            f"{action}_prompt": prompt,
            f"{action}_negative_prompt": negative_prompt,
            f"{action}_steps": self.steps,
            f"{action}_ddim_eta": self.ddim_eta,  # only applies to ddim scheduler
            f"{action}_n_iter": 1,
            f"{action}_width": self.width,
            f"{action}_height": self.height,
            f"{action}_n_samples": samples,
            f"{action}_scale": self.scale / 100,
            f"{action}_seed": seed,
            f"{action}_model": model,
            f"{action}_scheduler": self.scheduler,
            f"{action}_model_path": model_path,
            f"{action}_model_branch": model_branch,
            f"{action}_lora": self.available_lora(action),
            f"generator_section": self.currentTabSection,
            f"width": self.width,
            f"height": self.height,
            "do_nsfw_filter": self.settings_manager.settings.nsfw_filter.get(),
            "model_base_path": self.model_base_path,
            "pos_x": 0,
            "pos_y": 0,
            "outpaint_box_rect": self.active_rect,
            "hf_token": self.settings_manager.settings.hf_api_key.get(),
            "use_controlnet": use_controlnet,
            "controlnet": controlnet,
        }

        if action == "superresolution":
            options["original_image_width"] = self.canvas.current_active_image_data.image.width
            options["original_image_height"] = self.canvas.current_active_image_data.image.height

        if action in ["img2img", "depth2img", "pix2pix", "controlnet"]:
            options[f"{action}_strength"] = self.strength / 100.0

        if action == "pix2pix":
            options[f"pix2pix_image_guidance_scale"] = self.image_scale

        """
        Emitting generate_signal with options allows us to pass more options to the dict from
        modal windows such as the image interpolation window.
        """
        self.generate_signal.emit(options)

        memory_options = {
            "use_last_channels": self.settings_manager.settings.use_last_channels.get(),
            "use_enable_sequential_cpu_offload": self.settings_manager.settings.use_enable_sequential_cpu_offload.get(),
            "enable_model_cpu_offload": self.settings_manager.settings.enable_model_cpu_offload.get(),
            "use_attention_slicing": self.settings_manager.settings.use_attention_slicing.get(),
            "use_tf32": self.settings_manager.settings.use_tf32.get(),
            "use_cudnn_benchmark": self.settings_manager.settings.use_cudnn_benchmark.get(),
            "use_enable_vae_slicing": self.settings_manager.settings.use_enable_vae_slicing.get(),
            "use_accelerated_transformers": self.settings_manager.settings.use_accelerated_transformers.get(),
            "use_torch_compile": self.settings_manager.settings.use_torch_compile.get(),
            "use_tiled_vae": self.settings_manager.settings.use_tiled_vae.get(),
        }

        data = {
            "action": action,
            "options": {
                **options,
                **extra_options,
                **memory_options
            }
        }
        # data = self.do_generate_data_injection(data)  # TODO: Extensions
        self.client.message = data

    """
    TODO: Extensions
    def do_generate_data_injection(self, data):
        for extension in self.settings_manager.settings.active_extensions.get():
            data = extension.generate_data_injection(data)
        return data
    """

    def tab_changed_callback(self, index):
        self.set_final_size_label()
        self.canvas.update()

    def set_default_values(self, section, tab):
        tab.prompt.setPlainText(self.prompt)
        tab.negative_prompt.setPlainText(self.negative_prompt)
        tab.steps_spinbox.setValue(self.steps)
        tab.scale_spinbox.setValue(self.scale / 100)
        if section == "pix2pix":
            val = self.settings_manager.settings.pix2pix_image_guidance_scale.get()
            tab.image_scale_spinbox.setValue(val / 100)
            if type(val) == float:
                val = int(val * 100)
            tab.image_scale_slider.setValue(val)
        try:
            tab.strength_spinbox.setValue(self.strength / 100)
        except:
            pass
        tab.seed.setText(str(self.seed))
        tab.samples_spinbox.setValue(self.samples)
        tab.model_dropdown.setCurrentText(self.model)
        try:
            tab.scheduler_dropdown.setCurrentText(self.scheduler)
        except RuntimeError:
            pass

    def handle_width_slider_change(self, val):
        self.window.width_spinbox.setValue(val)
        self.width = val

    def handle_width_spinbox_change(self, val):
        self.window.width_slider.setValue(int(val))
        self.width = int(val)

    def handle_height_slider_change(self, val):
        self.window.height_spinbox.setValue(int(val))
        self.height = int(val)

    def handle_height_spinbox_change(self, val):
        self.window.height_slider.setValue(int(val))
        self.height = int(val)

    def update_brush_size(self, val):
        self.settings_manager.settings.mask_brush_size.set(val)
        self.window.brush_size_spinbox.setValue(val)

    def brush_spinbox_change(self, val):
        self.settings_manager.settings.mask_brush_size.set(val)
        self.window.brush_size_slider.setValue(val)

    def initialize_size_form_elements(self):
        # width form elements
        self.window.width_slider.valueChanged.connect(lambda val: self.handle_width_slider_change(val))
        self.window.width_spinbox.valueChanged.connect(lambda val: self.handle_width_spinbox_change(val))

        # height form elements
        self.window.height_slider.valueChanged.connect(lambda val: self.handle_height_slider_change(val))
        self.window.height_spinbox.valueChanged.connect(lambda val: self.handle_height_spinbox_change(val))

    def initialize_size_sliders(self):
        self.window.width_slider.setValue(self.width)
        self.window.height_slider.setValue(self.height)
        self.window.width_spinbox.setValue(self.width)
        self.window.height_spinbox.setValue(self.height)
        self.window.brush_size_slider.setValue(self.settings.mask_brush_size.get())
        self.window.brush_size_slider.valueChanged.connect(self.update_brush_size)
        self.window.brush_size_spinbox.valueChanged.connect(self.brush_spinbox_change)
        self.window.brush_size_slider.setValue(self.settings_manager.settings.mask_brush_size.get())
        self.window.brush_size_spinbox.setValue(self.settings_manager.settings.mask_brush_size.get())
        self.set_size_form_element_step_values()

    def load_template(self, template_name):
        try:
            return uic.loadUi(
                os.path.join("pyqt", f"{template_name}.ui"))
        except UIFileException:
            return None



from PyQt6.QtWidgets import QColorDialog


class BrushesMixin:
    def initialize(self):
        self.window.primary_color_button.clicked.connect(self.set_primary_color)
        self.set_button_colors()

    def reset_settings(self):
        self.window.primary_color_button.setStyleSheet(f"background-color: {self.settings_manager.settings.primary_color.get()};")
        self.window.brush_size_slider.setValue(self.settings_manager.settings.size.get())

    def set_primary_color(self):
        # display a color picker
        color = QColorDialog.getColor()
        if color.isValid():
            self.settings_manager.settings.primary_color.set(color.name())
            self.set_button_colors()

    def set_button_colors(self):
        self.window.primary_color_button.setStyleSheet(
            f"background-color: {self.settings_manager.settings.primary_color.get()};"
        )



from PIL import ImageDraw, ImageFilter, UnidentifiedImageError
from PyQt6.QtCore import Qt, QPointF, QPoint, QSize, QRect, QThread, QObject, pyqtSignal, QTimer, QRunnable, QThreadPool
from PyQt6.QtGui import QPainter, QPainterPath, QColor, QPen, QImage
from airunner.models.layerdata import LayerData
from airunner.models.linedata import LineData
from PIL import Image


class RasterizationWorker(QObject):
    finished = pyqtSignal()

    def __init__(self, *args, **kwargs):
        self.convert_pixmap_to_pil_image = kwargs.pop('convert_pixmap_to_pil_image')
        self.img = kwargs.pop('img')
        self.top = kwargs.pop('top')
        self.left = kwargs.pop('left')
        self.bottom = kwargs.pop('bottom')
        self.right = kwargs.pop('right')
        self.layer = kwargs.pop("layer")
        super().__init__(*args)

    def run(self):
        self.convert_pixmap_to_pil_image(
            self.img,
            self.top,
            self.left,
            self.bottom,
            self.right,
            self.layer
        )
        self.finished.emit()


class RasterizationTask(QRunnable):
    def __init__(self, worker):
        super().__init__()
        self.worker = worker

    def run(self):
        self.worker.run()


class CanvasBrushesMixin:
    thread = None
    worker = None

    @property
    def left_line_extremity(self):
        return self.current_layer.left_line_extremity

    @left_line_extremity.setter
    def left_line_extremity(self, value):
        self.current_layer.left_line_extremity = value

    @property
    def right_line_extremity(self):
        return self.current_layer.right_line_extremity

    @right_line_extremity.setter
    def right_line_extremity(self, value):
        self.current_layer.right_line_extremity = value

    @property
    def top_line_extremity(self):
        return self.current_layer.top_line_extremity

    @top_line_extremity.setter
    def top_line_extremity(self, value):
        self.current_layer.top_line_extremity = value

    @property
    def bottom_line_extremity(self):
        return self.current_layer.bottom_line_extremity

    @bottom_line_extremity.setter
    def bottom_line_extremity(self, value):
        self.current_layer.bottom_line_extremity = value

    @property
    def last_left(self):
        return self.current_layer.last_left

    @last_left.setter
    def last_left(self, value):
        self.current_layer.last_left = value

    @property
    def last_top(self):
        return self.current_layer.last_top

    @last_top.setter
    def last_top(self, value):
        self.current_layer.last_top = value

    @property
    def min_x(self):
        return self.current_layer.min_x

    @min_x.setter
    def min_x(self, value):
        self.current_layer.min_x = value

    @property
    def min_y(self):
        return self.current_layer.min_y

    @min_y.setter
    def min_y(self, value):
        self.current_layer.min_y = value

    @property
    def last_pos(self):
        return self.current_layer.last_pos

    @last_pos.setter
    def last_pos(self, value):
        self.current_layer.last_pos = value

    @property
    def color(self):
        return self.current_layer.color

    @color.setter
    def color(self, value):
        self.current_layer.color = value

    @property
    def line_width(self):
        return self.current_layer.line_width

    @line_width.setter
    def line_width(self, value):
        self.current_layer.line_width = value

    @property
    def opacity(self):
        return self.current_layer.opacity

    def initialize(self):
        self.thread_pool = QThreadPool()
        self.thread_pool.setMaxThreadCount(4)

    def draw(self, layer, index):
        path = QPainterPath()
        painter = None
        for line in layer.lines:
            pen = line.pen
            pen.setCapStyle(Qt.PenCapStyle.RoundCap)
            pen.setJoinStyle(Qt.PenJoinStyle.RoundJoin)
            # set opacity based on current_layer opacity
            if self.line_width is None or self.line_width != line.width:
                self.line_width = line.width
                self.draw_path(path, painter)
                painter = None
            if self.color is None or self.color != line.color:
                self.color = line.color
                self.draw_path(path, painter)
                painter = None
            if not painter:
                painter = QPainter(self.canvas_container)
                painter.setBrush(self.brush)
                painter.setOpacity(1.0)
                path = QPainterPath()
            painter.setPen(pen)

            start = QPointF(line.start_point.x() + self.pos_x, line.start_point.y() + self.pos_y)
            end = QPointF(line.end_point.x() + self.pos_x, line.end_point.y() + self.pos_y)

            # also apply the layer offset
            offset = QPointF(self.current_layer.offset.x(), self.current_layer.offset.y())
            start += offset
            end += offset

            # calculate control points for the Bezier curve
            dx = end.x() - start.x()
            dy = end.y() - start.y()
            ctrl1 = QPointF(start.x() + dx / 3, start.y() + dy / 3)
            ctrl2 = QPointF(end.x() - dx / 3, end.y() - dy / 3)

            # add the curve to the path
            path.moveTo(start)
            path.cubicTo(ctrl1, ctrl2, end)

        # draw the entire line with a single drawPath call
        self.draw_path(path, painter)

    def draw_path(self, path, painter):
        if painter:
            painter.drawPath(path)
            painter.end()

    def handle_erase(self, event):
        self.is_erasing = True
        if len(self.current_layer.lines) > 0:
            self.rasterize_lines(final=True)
        brush_size = int(self.settings_manager.settings.mask_brush_size.get() / 2)
        image = self.current_layer.image_data.image if self.current_layer.image_data.image is not None else None
        image_pos = self.current_layer.image_data.position if self.current_layer.image_data.image is not None else None
        if image is None:
            return
        start = event.pos() - QPoint(self.pos_x, self.pos_y) - image_pos
        if image:
            image = image.copy()
            draw = ImageDraw.Draw(image)
            if self.last_pos is None:
                self.last_pos = start
            draw.line([
                self.last_pos.x(),
                self.last_pos.y(),
                start.x(),
                start.y()
            ], fill=(0, 0, 0, 0), width=brush_size*2, joint="curve")
            draw.ellipse((
                start.x() - brush_size,
                start.y() - brush_size,
                start.x() + brush_size,
                start.y() + brush_size
            ), fill=(0, 0, 0, 0))
            self.current_layer.image_data.image = image
            self.last_pos = start
            self.update()
        self.update()

    def pen(self, event):
        brush_color = "#ffffff"
        if event.button() == Qt.MouseButton.LeftButton or Qt.MouseButton.LeftButton in event.buttons():
            brush_color = self.settings_manager.settings.primary_color.get()
        brush_color = QColor(brush_color)
        pen = QPen(
            brush_color,
            self.settings_manager.settings.mask_brush_size.get()
        )
        return pen

    def handle_draw(self, event):
        start = event.pos() - QPoint(self.pos_x, self.pos_y)
        pen = self.pen(event)
        if len(self.current_layer.lines) > 0:
            previous = LineData(self.current_layer.lines[-1].start_point, start, pen, self.current_layer_index)
            self.current_layer.lines[-1] = previous

        end = event.pos() - QPoint(self.pos_x + 1, self.pos_y)
        line_data = LineData(start, end, pen, self.current_layer_index)
        self.current_layer.lines.append(line_data)
        self.update()

    def get_line_extremities(self, lines):
        for line in lines:
            start_x = line.start_point.x()
            start_y = line.start_point.y()
            end_x = line.end_point.x()
            end_y = line.end_point.y()

            brush_size = int(self.settings_manager.settings.mask_brush_size.get() / 2)
            min_x = min(start_x, end_x) - brush_size
            min_y = min(start_y, end_y) - brush_size
            max_x = max(start_x, end_x) + brush_size
            max_y = max(start_y, end_y) + brush_size
            if self.left_line_extremity is None:
                self.left_line_extremity = min_x
            else:
                self.left_line_extremity = min(self.left_line_extremity, min_x)
            if self.right_line_extremity is None:
                self.right_line_extremity = max_x
            else:
                self.right_line_extremity = max(self.right_line_extremity, max_x)
            if self.top_line_extremity is None:
                self.top_line_extremity = min_y
            else:
                self.top_line_extremity = min(self.top_line_extremity, min_y)
            if self.bottom_line_extremity is None:
                self.bottom_line_extremity = max_y
            else:
                self.bottom_line_extremity = max(self.bottom_line_extremity, max_y)
        return self.top_line_extremity, self.left_line_extremity, self.bottom_line_extremity, self.right_line_extremity

    def rasterize_lines(self, final=False):
        max_lines = len(self.current_layer.lines)
        if max_lines == 0:
            return

        lines = self.current_layer.lines[:max_lines]
        top, left, bottom, right = self.get_line_extremities(lines)
        brush_size = self.settings_manager.settings.mask_brush_size.get()
        if brush_size > 1:
            brush_size = int(brush_size / 2)

        # create a QImage with the size of the lines
        min_x = min(left, right) - brush_size
        max_x = max(left, right) - brush_size
        min_y = min(top, bottom) - brush_size
        max_y = max(top, bottom) - brush_size
        width = abs(max_x - min_x)
        height = abs(max_y - min_y)
        img = QImage(QSize(width, height), QImage.Format.Format_ARGB32)
        img.fill(Qt.GlobalColor.transparent)
        painter = QPainter(img)
        painter.setBrush(self.brush)
        painter.setOpacity(1.0)
        path = self.create_image_path(painter, lines)
        painter.drawPath(path)
        painter.end()

        worker = RasterizationWorker(
            convert_pixmap_to_pil_image=self.convert_pixmap_to_pil_image,
            img=img,
            top=top,
            left=left,
            bottom=bottom,
            right=right,
            layer=self.current_layer
        )
        task = RasterizationTask(worker)
        self.thread_pool.start(task)
        # task.setAutoDelete(True)
        task.worker.finished.connect(lambda _max_lines=max_lines, _final=final: self.finalize_pixmap(_max_lines, _final))

    def finalize_pixmap(self, max_lines, final=False):
        self.current_layer.lines = self.current_layer.lines[max_lines:]
        self.update()

    def create_image_path(self, painter, lines):
        path = QPainterPath()
        path.setFillRule(Qt.FillRule.WindingFill)
        for line in lines:
            pen = line.pen
            color = line.color
            pen.setColor(color)
            pen.setCapStyle(Qt.PenCapStyle.RoundCap)
            pen.setJoinStyle(Qt.PenJoinStyle.RoundJoin)
            painter.setPen(pen)

            start = QPointF(line.start_point.x() - self.left_line_extremity, line.start_point.y() - self.top_line_extremity)
            end = QPointF(line.end_point.x() - self.left_line_extremity, line.end_point.y() - self.top_line_extremity)

            # also apply the layer offset
            offset = QPointF(self.current_layer.offset.x(), self.current_layer.offset.y())
            start += offset
            end += offset

            # calculate control points for the Bezier curve
            dx = end.x() - start.x()
            dy = end.y() - start.y()
            ctrl1 = QPointF(start.x() + dx / 3, start.y() + dy / 3)
            ctrl2 = QPointF(end.x() - dx / 3, end.y() - dy / 3)

            # add the curve to the path
            path.moveTo(start)
            path.cubicTo(ctrl1, ctrl2, end)
        return path

    def convert_pixmap_to_pil_image(
        self,
        img: Image,
        top: int,
        left: int,
        bottom: int,
        right: int,
        layer: LayerData
    ):
        try:
            img = Image.fromqpixmap(img)
        except UnidentifiedImageError:
            return None
        self.insert_rasterized_line_image(
            QRect(left, top, right, bottom),
            img,
            layer
        )



from PyQt6.QtGui import QPainter


class CanvasWidgetsMixin:
    def draw(self, layer, index):
        painter = QPainter(self.canvas_container)
        self.draw_widgets(layer, index, painter)
        painter.end()

    def draw_widgets(self, layer, index, painter):
        for widget in layer.widgets:
            widget.draw(painter)



import io
import random
import subprocess
from PIL import Image, ImageGrab, ImageOps
from PIL.ImageQt import ImageQt
from PyQt6.QtCore import QPoint, QRect
from PyQt6.QtGui import QPainter, QPixmap
from PIL.ImageFilter import GaussianBlur
from airunner.filters.rgb_noise_filter import RGBNoiseFilter
from airunner.models.imagedata import ImageData
from PIL.ExifTags import TAGS
from airunner.models.layerdata import LayerData


class CanvasImageMixin:
    _image_data_copy = None

    @property
    def image_data(self):
        if self._image_data_copy is None:
            self._image_data_copy = self.image_data_copy(self.current_layer_index)
        return self._image_data_copy
    
    @image_data.setter
    def image_data(self, value):
        self._image_data_copy = value

    @property
    def current_active_image_data(self):
        try:
            return self.current_layer.image_data
        except IndexError:
            return None

    def initialize(self):
        pass

    def apply_filter(self, filter):
        if self.current_layer.image_data.image is None:
            return
        self.parent.history.add_event({
            "event": "apply_filter",
            "layer_index": self.current_layer_index,
            "images": self.image_data,
        })

        if type(filter).__name__ in ["SaturationFilter", "ColorBalanceFilter", "RGBNoiseFilter", "PixelFilter"]:
            filtered_image = filter.filter(self.image_data.image)
        else:
            filtered_image = self.image_data.image.filter(filter)
        self.current_layer.image_data.image = filtered_image
        self.image_data = None

    def preview_filter(self, filter):
        if self.current_layer.image_data.image is None:
            return
        # check if filter is a SaturationFilter object
        if type(filter).__name__ in ["SaturationFilter", "ColorBalanceFilter", "RGBNoiseFilter", "PixelFilter"]:
            filtered_image = filter.filter(self.image_data.image)
        else:
            filtered_image = self.image_data.image.filter(filter)
        self.current_layer.image_data.image = filtered_image

    def cancel_filter(self):
        self.current_layer.image_data = self.image_data
        self.image_data = None

    def draw(self, layer, index):
        painter = QPainter(self.canvas_container)
        self.draw_images(layer, painter)
        painter.end()

    def visible_image(self, layer=None, image_data=None):
        """
        Returns an image that is cropped to the visible area of the canvas
        :param layer:
        :return: Image
        """
        if layer:
            image_data = layer.image_data

        if not image_data:
            return

        img = image_data.image
        if not img:
            return

        # apply the layer offset
        x = image_data.position.x() + self.pos_x
        y = image_data.position.y() + self.pos_y
        location = QPoint(int(x), int(y))# + layer.offset

        rect = self.viewport_rect

        # only create a image of the visible area, apply offset
        img = img.copy().crop((
            rect.x() - location.x(),
            rect.y() - location.y(),
            rect.x() + rect.width() - location.x(),
            rect.y() + rect.height() - location.y()
        ))
        return img

    def draw_images(self, layer, painter):
        img = self.visible_image(layer=layer)
        if img:
            qimage = ImageQt(img)
            pixmap = QPixmap.fromImage(qimage)
            painter.drawPixmap(QPoint(0, 0), pixmap)

    def copy_image(self):
        im = self.current_active_image_data
        if not im:
            return
        output = io.BytesIO()
        if self.parent.is_windows:
            im.image.save(output, format="DIB")
            self.image_to_system_clipboard_windows(output.getvalue())
        else:
            im.image.save(output, format="PNG")
            self.image_to_system_clipboard_linux(output.getvalue())

    def image_to_system_clipboard_windows(self, data):
        import win32clipboard
        win32clipboard.OpenClipboard()
        win32clipboard.EmptyClipboard()
        win32clipboard.SetClipboardData(win32clipboard.CF_DIB, data)
        win32clipboard.CloseClipboard()

    def image_to_system_clipboard_linux(self, data):
        try:
            subprocess.Popen(["xclip", "-selection", "clipboard", "-t", "image/png"],
                             stdin=subprocess.PIPE).communicate(data)
        except FileNotFoundError:
            pass

    def image_to_system_clipboard_windows(self, data):
        import win32clipboard
        win32clipboard.OpenClipboard()
        win32clipboard.EmptyClipboard()
        win32clipboard.SetClipboardData(win32clipboard.CF_DIB, data)
        win32clipboard.CloseClipboard()

    def image_from_system_clipboard_windows(self):
        import win32clipboard
        try:
            win32clipboard.OpenClipboard()
            data = win32clipboard.GetClipboardData(win32clipboard.CF_DIB)
            win32clipboard.CloseClipboard()
            # convert bytes to image
            image = Image.open(io.BytesIO(data))
            return image
        except Exception as e:
            #self.parent.error_handler(str(e))
            print(e)
            return None

    def image_from_system_clipboard_linux(self):
        try:
            image = ImageGrab.grabclipboard()
            # with transparency
            image = image.convert("RGBA")
            return image
        except Exception as e:
            return None

    def paste_image_from_clipboard(self):
        if self.parent.is_windows:
            image = self.image_from_system_clipboard_windows()
        else:
            image = self.image_from_system_clipboard_linux()

        if not image:
            return

        if self.settings_manager.settings.resize_on_paste.get():
            if self.settings_manager.settings.resize_on_paste.get():
                image.thumbnail((self.settings_manager.settings.working_width.get(),
                                 self.settings_manager.settings.working_height.get()), Image.ANTIALIAS)
        self.create_image(QPoint(0, 0), image)
        self.update()

    def create_image(self, location, image):
        """
        Create a new image object and add it to the current layer
        """
        # convert image to RGBA
        image = image.convert("RGBA")
        self.current_layer.image_data = ImageData(
            position=location,
            image=image,
            opacity=self.current_layer.opacity
        )
        #self.set_image_opacity(self.get_layer_opacity(self.current_layer_index))

    def invert_image(self):
        # convert image mode to RGBA
        image_data = self.current_layer.image_data
        if image_data.image is not None:
            r, g, b, a = image_data.image.split()
            r = ImageOps.invert(r)
            g = ImageOps.invert(g)
            b = ImageOps.invert(b)
            image_data.image = Image.merge("RGBA", (r, g, b, a))

    def film_filter(self):
        working_images = self.parent.canvas.current_active_image_data
        if working_images is not None:
            self.parent.history.add_event({
                "event": "apply_filter",
                "layer_index": self.current_layer_index,
                "images": self.image_data,
            })
            image = working_images.image.copy()
            red_grain = Image.new("L", image.size)
            green_grain = Image.new("L", image.size)
            blue_grain = Image.new("L", image.size)
            red_grain.putdata([random.randint(0, 255) for _i in range(image.size[0] * image.size[1])])
            green_grain.putdata([random.randint(0, 255) for _i in range(image.size[0] * image.size[1])])
            blue_grain.putdata([random.randint(0, 255) for _i in range(image.size[0] * image.size[1])])
            filter = RGBNoiseFilter(0.1, 0.1, 0.1, red_grain, green_grain, blue_grain)
            filtered_image = filter.filter(self.image_data.image)
            filtered_image = filtered_image.filter(GaussianBlur(radius=0.03 * 20))
            self.current_layer.image_data.image = filtered_image
            self.image_data = None
        self.update()

    def load_image(self, image_path):
        image = Image.open(image_path)
        self.load_metadata(image)

        # if settings_manager.settings.resize_on_paste, resize the image to working width and height while mainting its aspect ratio
        if self.settings_manager.settings.resize_on_paste.get():
            image.thumbnail((self.settings_manager.settings.working_width.get(), self.settings_manager.settings.working_height.get()), Image.ANTIALIAS)

        self.create_image(QPoint(0, 0), image)
        self.update()

    def load_metadata(self, image):
        if not self.settings_manager.settings.import_metadata.get():
            return
        try:
            metadata = image.text
        except AttributeError:
            metadata = None
        self.parent.load_metadata(metadata)

    def save_image(self, image_path, image=None):
        if self.current_layer.image_data.image is None:
            return
        if image is None:
            image = self.current_layer.image_data.image
        image = image.convert("RGBA")
        if not "." in image_path:
            image_path += ".png"
        if not image_path.endswith(".png") and not image_path.endswith(".gif"):
            image = image.convert("RGB")
        image.save(image_path)
        self.saving = False
        self.update()

    def image_handler(self, active_img, data):
        self.update_image_canvas(data["action"], data, active_img)
        self.current_layer.lines = []
        self.update()

    def update_image_canvas(self, section=None, data=None, processed_image: Image = None):
        """
        Update the image by section
        This is used by other classes to add an image to the canvas
        :param section: the section (action) that was taken to generate this image. section is a deprecated name
        :param data: the data to update the image with
        :param processed_image: the image to update the canvas with
        :return:
        """
        processed_image = processed_image.convert("RGBA")
        section = data["action"] if not section else section
        outpaint_box_rect = data["options"]["outpaint_box_rect"]
        if section not in["superresolution", "upscale"]:
            processed_image, image_root_point, image_pivot_point = self.handle_outpaint(
                outpaint_box_rect,
                processed_image,
                section,
                is_kandinsky=data["options"].get("generator_section") == "kandinsky"
            )
        else:
            # if we are upscaling (or using superresolution) we want to replace the existing image with the new
            # one so that there is no undesired overlapping of images.
            # to do this, the handle_outpaint function is skipped and we reset the coordinates to 0,0
            image_root_point = QPoint(0, 0)
            image_pivot_point = QPoint(0, 0)
            self.pos_x = 0
            self.pos_y = 0

        data = {
            "processed_image": processed_image,
            "image_root_point": image_root_point,
            "image_pivot_point": image_pivot_point,
            "add_image_to_canvas": True
        }

        """
        We pass data through an emitter so that it can be modified by other classes and adding to canvas
        can be interrupted if needed. For example, the image interpolation window takes the processed image
        and displays it in the window rather than on the canvas.
        """
        self.parent.add_image_to_canvas_signal.emit(data)

        if data["add_image_to_canvas"]:
            self.add_image_to_canvas(
                data["processed_image"],
                image_root_point=data["image_root_point"],
                image_pivot_point=data["image_pivot_point"]
            )

    def insert_rasterized_line_image(self, rect: QRect, img: Image, layer: LayerData):
        existing_image = layer.image_data.image
        # combine img with existing image
        pos = [0, 0]
        if existing_image is None:
            point = QPoint(rect.x(), rect.y())
            width = img.width
            height = img.height
            new_image = Image.new("RGBA", (width, height), (0, 0, 0, 0))
        else:
            point = QPoint(
                min(rect.x(), layer.image_data.position.x()),
                min(rect.y(), layer.image_data.position.y())
            )
            if point.x() < rect.x():
                pos[0] = abs(point.x()) + rect.x()
            if point.y() < rect.y():
                pos[1] = abs(point.y()) + rect.y()
            existing_position = layer.image_data.position
            is_drawing_left = rect.x() < existing_position.x()
            is_drawing_right = rect.width() > (existing_image.width + existing_position.x())
            is_drawing_top = rect.y() < existing_position.y()
            is_drawing_bottom = rect.height() > (existing_image.height + existing_position.y())

            x_overlap = 0
            if is_drawing_left:
                x_overlap = existing_position.x() - rect.x()
            if is_drawing_right:
                x_overlap += rect.width() - (existing_image.width + existing_position.x())
            y_overlap = 0
            if is_drawing_top:
                y_overlap = existing_position.y() - rect.y()
            if is_drawing_bottom:
                y_overlap += rect.height() - (existing_image.height + existing_position.y())

            if is_drawing_left or is_drawing_right:
                width = existing_image.width + x_overlap
            else:
                width = max(existing_image.width, img.width)
            if is_drawing_top or is_drawing_bottom:
                height = existing_image.height + y_overlap
            else:
                height = max(existing_image.height, img.height)
            new_image = Image.new("RGBA", (width, height), (0, 0, 0, 0))

        if existing_image:
            pos_x = layer.image_data.position.x() - rect.x() if rect.x() < layer.image_data.position.x() else 0
            pos_y = layer.image_data.position.y() - rect.y() if rect.y() < layer.image_data.position.y() else 0
            new_image.alpha_composite(existing_image, (pos_x, pos_y))
        new_image.alpha_composite(img, (pos[0], pos[1]))
        self.add_image_to_canvas(
            new_image,
            image_root_point=point,
            image_pivot_point=point,
            layer=layer
        )

    def handle_outpaint(self, outpaint_box_rect, outpainted_image, action=None, is_kandinsky=False):
        if self.current_layer.image_data.image is None:
            point = QPoint(outpaint_box_rect.x(), outpaint_box_rect.y())
            return outpainted_image, self.image_root_point, point

        # make a copy of the current canvas image
        existing_image_copy = self.current_layer.image_data.image.copy()
        width = existing_image_copy.width
        height = existing_image_copy.height

        pivot_point = self.image_pivot_point
        root_point = self.image_root_point

        is_drawing_left = outpaint_box_rect.x() < pivot_point.x()
        is_drawing_right = outpaint_box_rect.width() > (width - abs(self.current_layer.image_data.position.x()))
        is_drawing_up = outpaint_box_rect.y() < pivot_point.y()
        is_drawing_down = outpaint_box_rect.height() > (height - abs(self.current_layer.image_data.position.y()))

        x_overlap = 0
        if is_drawing_left:
            x_overlap = pivot_point.x() - outpaint_box_rect.x()
        if is_drawing_right:
            x_overlap += outpaint_box_rect.width() - (width - abs(self.current_layer.image_data.position.x()))

        y_overlap = 0
        if is_drawing_up:
            y_overlap = pivot_point.y() - outpaint_box_rect.y()
        if is_drawing_down:
            y_overlap += outpaint_box_rect.height() - (height - abs(self.current_layer.image_data.position.y()))

        if is_drawing_left or is_drawing_right:
            x = width + x_overlap
        else:
            x = width
        if is_drawing_up or is_drawing_down:
            y = height + y_overlap
        else:
            y = height
        new_dimensions = (int(x), int(y))
        if new_dimensions[0] < width:
            new_dimensions = (width, new_dimensions[1])
        if new_dimensions[1] < height:
            new_dimensions = (new_dimensions[0], height)
        new_image = Image.new("RGBA", new_dimensions, (0, 0, 0, 0))
        new_image_a = Image.new("RGBA", new_dimensions, (0, 0, 0, 0))
        new_image_b = Image.new("RGBA", new_dimensions, (0, 0, 0, 0))
        existing_image_pos = [0, 0]
        image_root_point = QPoint(root_point.x(), root_point.y())
        image_pivot_point = QPoint(pivot_point.x(), pivot_point.y())
        if is_drawing_left:
            current_x_pos = abs(outpaint_box_rect.x() - image_pivot_point.x())
            left_overlap = abs(outpaint_box_rect.x()) - abs(image_root_point.x())
            image_root_point.setX(width + left_overlap)
            image_pivot_point.setX(int(outpaint_box_rect.x()))
            existing_image_pos = [current_x_pos, existing_image_pos[1]]
            pos_x = max(0, outpaint_box_rect.x() + pivot_point.x())
        else:
            pos_x = max(0, outpaint_box_rect.x() - pivot_point.x())
        if is_drawing_up:
            current_y_pos = abs(outpaint_box_rect.y() - image_pivot_point.y())
            up_overlap = abs(outpaint_box_rect.y()) - abs(image_root_point.y())
            image_root_point.setY(height + up_overlap)
            image_pivot_point.setY(int(outpaint_box_rect.y()))
            existing_image_pos = [existing_image_pos[0], current_y_pos]
            pos_y = max(0, outpaint_box_rect.y() + pivot_point.y())
        else:
            pos_y = max(0, outpaint_box_rect.y() - pivot_point.y())

        new_image_a.paste(outpainted_image, (int(pos_x), int(pos_y)))
        new_image_b.paste(existing_image_copy, (int(existing_image_pos[0]), int(existing_image_pos[1])))

        if action == "outpaint" and not is_kandinsky:
            new_image = Image.alpha_composite(new_image, new_image_a)
            new_image = Image.alpha_composite(new_image, new_image_b)
        else:
            new_image = Image.alpha_composite(new_image, new_image_b)
            new_image = Image.alpha_composite(new_image, new_image_a)

        return new_image, image_root_point, image_pivot_point

    def add_image_to_canvas(self, image, image_root_point, image_pivot_point, layer:LayerData=None):
        self.parent.history.add_event({
            "event": "set_image",
            "layer_index": self.current_layer_index,
            "images": self.current_layer.image_data
        })
        image = self.apply_opacity(image, self.current_layer.opacity)

        if not layer:
            layer = self.current_layer

        layer.image_data = ImageData(
            position=image_pivot_point,
            image=image,
            opacity=self.current_layer.opacity,
            image_root_point=image_root_point,
            image_pivot_point=image_pivot_point
        )

    def lower_opacity(self, i, diff):
        if i == 0:
            return 0
        total = i + diff
        return total if total > 0 else i

    def raise_opacity(self, i, diff):
        if i == 0:
            return 0
        total = i + diff
        return total

    def apply_opacity(self, image, target_opacity):
        if not image:
            return image
        target_opacity = 255 * target_opacity
        if target_opacity == 0:
            target_opacity = 1
        r, g, b, a = image.split()
        a = a.point(lambda i: target_opacity if i > 0 else 0)
        image.putalpha(a)
        return image

    def image_data_copy(self, index):
        image_data = self.layers[index].image_data
        return ImageData(
            position=image_data.position,
            image=image_data.image.copy() if image_data.image else None,
            opacity=self.current_layer.opacity,
            image_pivot_point=image_data.image_pivot_point,
            image_root_point=image_data.image_root_point
        )

    def rotate_90_clockwise(self):
        if self.current_active_image_data:
            if not self.current_active_image_data.image:
                return
            self.parent.history.add_event({
                "event": "rotate",
                "layer_index": self.current_layer_index,
                "images": self.image_data_copy(self.current_layer_index)
            })
            if self.current_active_image_data.image:
                self.current_active_image_data.image = self.current_active_image_data.image.transpose(Image.ROTATE_270)
            self.update()

    def rotate_90_counterclockwise(self):
        if self.current_active_image_data:
            if not self.current_active_image_data or not self.current_active_image_data.image:
                return
            self.parent.history.add_event({
                "event": "rotate",
                "layer_index": self.current_layer_index,
                "images": self.image_data_copy(self.current_layer_index)
            })
            if self.current_active_image_data.image:
                self.current_active_image_data.image = self.current_active_image_data.image.transpose(Image.ROTATE_90)
            self.update()



from PyQt6.QtCore import Qt, QPoint, QRect
from PyQt6.QtGui import QColor, QPainter, QBrush, QCursor
from airunner.cursors.circle_brush import CircleCursor
from airunner.mixins.canvas_active_grid_area_mixin import CanvasActiveGridAreaMixin
from airunner.mixins.canvas_brushes_mixin import CanvasBrushesMixin
from airunner.mixins.canvas_grid_mixin import CanvasGridMixin
from airunner.mixins.canvas_image_mixin import CanvasImageMixin
from airunner.mixins.canvas_layer_mixin import CanvasLayerMixin
from airunner.mixins.canvas_selectionbox_mixin import CanvasSelectionboxMixin
from airunner.mixins.canvas_widgets_mixin import CanvasWidgetsMixin
from airunner.models.linedata import LineData


class Canvas(
    CanvasBrushesMixin,
    CanvasGridMixin,
    CanvasWidgetsMixin,
    CanvasImageMixin,
    CanvasSelectionboxMixin,
    CanvasActiveGridAreaMixin,
    CanvasLayerMixin,
):
    saving = False
    select_start = None
    select_end = None
    shift_is_pressed = False
    left_mouse_button_down = False
    brush_start = None
    last_mouse_pos = None

    @property
    def image_pivot_point(self):
        return self.current_layer.image_data.image_pivot_point

    @image_pivot_point.setter
    def image_pivot_point(self, value):
        self.current_layer.image_data.image_pivot_point = value

    @property
    def image_root_point(self):
        return self.current_layer.image_data.image_root_point

    @image_root_point.setter
    def image_root_point(self, value):
        self.current_layer.image_data.image_root_point = value

    @property
    def current_layer(self):
        if len(self.layers) == 0:
            return None
        return self.layers[self.current_layer_index]

    @property
    def select_selected(self):
        return self.settings_manager.settings.current_tool.get() == "select"

    @property
    def eraser_selected(self):
        return self.settings_manager.settings.current_tool.get() == "eraser"

    @property
    def brush_selected(self):
        return self.settings_manager.settings.current_tool.get() == "brush"

    @property
    def move_selected(self):
        return self.settings_manager.settings.current_tool.get() == "move"

    @property
    def is_dragging(self):
        return False

    @property
    def is_zooming(self):
        return False

    @property
    def mouse_pos(self):
        return self.canvas_container.mapFromGlobal(QCursor.pos())

    @property
    def brush_size(self):
        return self.settings_manager.settings.mask_brush_size.get()

    @property
    def canvas_container(self):
        return self.parent.window.canvas_container

    @property
    def settings_manager(self):
        return self.parent.settings_manager

    @property
    def mouse_position(self):
        return self.canvas_container.mapFromGlobal(QCursor.pos())

    @property
    def is_drawing(self):
        return self.left_mouse_button_down and self.brush_selected

    @property
    def primary_color(self):
        return QColor(self.settings_manager.settings.primary_color.get())

    @property
    def viewport_rect(self):
        rect = self.canvas_container.contentsRect()
        rect = QRect(0, 0, rect.width(), rect.height())
        return rect

    def get_layer_opacity(self, index):
        return self.layers[index].opacity

    def set_layer_opacity(self, index, opacity):
        layer = self.layers[index]
        layer.opacity = opacity
        self.update()
        layer.image_data.image = self.apply_opacity(layer.image_data.image, opacity)

    def __init__(
        self,
        parent=None
    ):
        self.canvas_rect = QRect(0, 0, 0, 0)
        self.pos_x = 0
        self.pos_y = 0
        self.current_layer_index = 0
        self.is_erasing = False
        self.parent = parent

        CanvasGridMixin.initialize(self)
        CanvasActiveGridAreaMixin.initialize(self)
        CanvasLayerMixin.initialize(self)
        CanvasBrushesMixin.initialize(self)
        CanvasImageMixin.initialize(self)

        # Set initial position and size of the canvas
        self.canvas_container.setGeometry(QRect(
            int(self.canvas_rect.x()),
            int(self.canvas_rect.y()),
            int(self.canvas_rect.width()),
            int(self.canvas_rect.height())
        ))

        # set self.parent paintEvent
        self.canvas_container.paintEvent = self.paintEvent
        self.canvas_container.mousePressEvent = self.mouse_press_event
        self.canvas_container.mouseMoveEvent = self.mouse_move_event
        self.canvas_container.mouseReleaseEvent = self.mouse_release_event

        # on shift down
        self.parent.window.keyPressEvent = self.key_press_event

        # on key up
        self.parent.window.keyReleaseEvent = self.key_release_event

        # on mouse hover
        self.canvas_container.enterEvent = self.enter_event
        self.canvas_container.leaveEvent = self.leave_event

        # Set the default brush color for drawing
        self.brush = QBrush()
        self.brush.setStyle(Qt.BrushStyle.SolidPattern)

        # Set the initial position for mouse dragging
        self.drag_pos = QPoint(0, 0)

        self.set_canvas_color()

    def timerEvent(self, event):
        pass

    def key_press_event(self, event):
        if event.key() == Qt.Key.Key_Shift:
            self.shift_is_pressed = True

    def key_release_event(self, event):
        if event.key() == Qt.Key.Key_Shift:
            self.shift_is_pressed = False

    def set_canvas_color(self):
        self.canvas_container.setStyleSheet(f"background-color: {self.settings_manager.settings.canvas_color.get()};")
        self.canvas_container.setAutoFillBackground(True)

    def paintEvent(self, event):
        CanvasGridMixin.paintEvent(self, event)
        layers = self.layers.copy()
        layers.reverse()
        for index in range(len(layers)):
            layer = layers[index]
            if not layer.visible:
                continue
            if self.settings_manager.settings.show_active_image_area.get():
                self.draw_active_image_area()
            CanvasImageMixin.draw(self, layer, index)
            CanvasBrushesMixin.draw(self, layer, index)
            CanvasWidgetsMixin.draw(self, layer, index)
        CanvasSelectionboxMixin.paint_event(self, event)
        CanvasActiveGridAreaMixin.paint_event(self, event)

    def draw_active_image_area(self):
        """
        Draw a border around the active image area and fill the background with white
        :return:
        """
        # first iterate over all layers and get extremities
        left = None
        top = None
        right = None
        bottom = None
        for layer in self.layers:
            image_data = layer.image_data
            if image_data.image is None:
                continue
            left = min(left, image_data.position.x()) if left is not None else image_data.position.x()
            top = min(top, image_data.position.y()) if top is not None else image_data.position.y()
            right = max(right, image_data.image.width) if right is not None else image_data.image.width
            bottom = max(bottom, image_data.image.height) if bottom is not None else image_data.image.height
        if left is None or top is None or right is None or bottom is None:
            return

        painter = QPainter(self.canvas_container)
        color = QColor(255, 255, 255, 128)
        painter.setPen(color)
        painter.setBrush(color)
        painter.drawRect(self.pos_x + left, self.pos_y + top, right, bottom)
        painter.end()

    def enter_event(self, event):
        self.update_cursor()

    def update_cursor(self):
        if self.brush_selected or self.eraser_selected:
            self.canvas_container.setCursor(CircleCursor(Qt.GlobalColor.white, Qt.GlobalColor.transparent, self.brush_size))
        elif self.move_selected:
            self.canvas_container.setCursor(Qt.CursorShape.OpenHandCursor)
        elif self.active_grid_area_selected:
            self.canvas_container.setCursor(Qt.CursorShape.DragMoveCursor)
        else:
            self.canvas_container.setCursor(QCursor(Qt.CursorShape.ArrowCursor))

    def leave_event(self, event):
        self.canvas_container.setCursor(QCursor(Qt.CursorShape.ArrowCursor))

    def update(self):
        self.parent.window.canvas_position.setText(f"X: {self.pos_x}, Y: {self.pos_y}")

        self.canvas_container.update(self.viewport_rect)

    def clear(self):
        self.current_layer.lines = []
        self.current_layer.image_data = None
        self.update()

    def recenter(self):
        self.pos_x = 0
        self.pos_y = 0
        self.update()

    def handle_move_canvas(self, event):
        self.pos_x += event.pos().x() - self.drag_pos.x()
        self.pos_y += event.pos().y() - self.drag_pos.y()
        self.drag_pos = event.pos()
        self.update()

    def handle_move_layer(self, event):
        point = QPoint(
            event.pos().x() if self.drag_pos is not None else 0,
            event.pos().y() if self.drag_pos is not None else 0
        )
        # snap to grid
        grid_size = self.settings_manager.settings.size.get()
        point.setX(point.x() - (point.x() % grid_size))
        point.setY(point.y() - (point.y() % grid_size))

        # center the image
        # point.setX(int((point.x() - self.current_layer.images[0].image.size[0] / 2)))
        # point.setY(int((point.y() - self.current_layer.images[0].image.size[1] / 2)))

        # establish a rect based on line points - we need the area that is being moved
        # so that we can center the point on it
        rect = QRect()
        for line in self.current_layer.lines:
            rect = rect.united(QRect(line.start_point, line.end_point))

        try:
            rect = rect.united(QRect(
                self.current_layer.image_data.position.x(),
                self.current_layer.image_data.position.y(),
                self.current_layer.image_data.image.size[0],
                self.current_layer.image_data.image.size[1]
            ))
        except IndexError:
            pass

        # center the point on the rect
        point.setX(int(point.x() - int(rect.width() / 2)))
        point.setY(int(point.y() - int(rect.height() / 2)))

        self.layers[self.current_layer_index].offset = point

    def mouse_press_event(self, event):
        if event.button() == Qt.MouseButton.LeftButton:
            self.select_start = event.pos()
            if self.brush_selected:
                self.last_mouse_pos = event.pos()
                start = event.pos() - QPoint(self.pos_x, self.pos_y)
                end = event.pos() - QPoint(self.pos_x, self.pos_y)
                pen = self.pen(event)
                line = LineData(start, end, pen, self.current_layer_index)
                start += self.layers[self.current_layer_index].offset
                end += self.layers[self.current_layer_index].offset
                self.current_layer.lines += [line]

                if event.button() == Qt.MouseButton.LeftButton and not self.left_mouse_button_down:
                    self.left_mouse_button_down = True
                    self.brush_start = start

            self.handle_tool(event)
            self.update()
        elif event.button() == Qt.MouseButton.MiddleButton:
            # Start dragging the canvas when the middle or right mouse button is pressed
            self.drag_pos = event.pos()

    def mouse_move_event(self, event):
        # check if LeftButton is pressed
        if Qt.MouseButton.LeftButton in event.buttons():
            self.last_mouse_pos = event.pos()
            self.handle_tool(event)
            self.update()
        elif Qt.MouseButton.MiddleButton in event.buttons() and self.drag_pos is not None:
            self.handle_move_canvas(event)
    
    def mouse_release_event(self, event):
        if event.button() == Qt.MouseButton.LeftButton:
            self.left_mouse_button_down = False
            if self.eraser_selected:
                self.last_pos = None
                self.is_erasing = False
            elif self.brush_selected:
                self.rasterize_lines(final=True)
        elif event.button() == Qt.MouseButton.MiddleButton:
            # Start dragging the canvas when the middle or right mouse button is pressed
            self.drag_pos = event.pos()

    def handle_select(self, event):
        if self.select_selected:
            if self.select_start is None:
                self.select_start = event.pos()
            else:
                self.select_end = event.pos()

        # snap to grid if enabled
        if self.settings_manager.settings.snap_to_grid.get():
            grid_size = self.settings_manager.settings.size.get()
            self.select_start.setX(self.select_start.x() - (self.select_start.x() % grid_size))
            self.select_start.setY(self.select_start.y() - (self.select_start.y() % grid_size))
            self.select_end.setX(self.select_end.x() - (self.select_end.x() % grid_size))
            self.select_end.setY(self.select_end.y() - (self.select_end.y() % grid_size))

        self.update()

    def handle_tool(self, event):
        if self.eraser_selected:
            if not self.is_erasing:
                self.parent.history.add_event({
                    "event": "erase",
                    "layer_index": self.current_layer_index,
                    "lines": self.current_layer.lines.copy(),
                    "images": self.image_data_copy(self.current_layer_index)
                })
            self.handle_erase(event)
            self.parent.is_dirty = True
        elif self.brush_selected:
            self.handle_draw(event)
            self.parent.is_dirty = True
        elif self.move_selected:
            self.handle_move_layer(event)
            self.parent.is_dirty = True
        elif self.select_selected:
            self.handle_select(event)
        elif self.active_grid_area_selected:
            self.handle_move_active_grid_area(event)

    def handle_move_active_grid_area(self, event):
        pos = event.pos()
        point = QPoint(
            pos.x(),
            pos.y()
        )

        # drag from the center of active_grid_area_pivot_point based on the size
        width = self.settings_manager.settings.working_width.get()
        height = self.settings_manager.settings.working_height.get()
        point -= QPoint(
            int((width / 2) + self.pos_x),
            int((height / 2) + self.pos_y)
        )

        if self.settings_manager.settings.snap_to_grid.get():
            point = QPoint(
                point.x() - (point.x() % self.grid_size),
                point.y() - (point.y() % self.grid_size)
            )

        self.active_grid_area_pivot_point = point

        # trigger draw event
        self.update()

    def reset_settings(self):
        self.parent.window.width_slider.setValue(self.settings_manager.settings.working_width.get())
        self.parent.window.height_slider.setValue(self.settings_manager.settings.working_height.get())
